//Underload to Scala compiler by Lymia
//Released to the public domain

package lymia

object Ulc {
  val VERSION = "v1.0"
  val USAGE = "Usage: [input file] > [output file]"
  val err = System.err
  def main(args: Array[String]): Unit = {
    err.println("Underload to Scala Compiler "+VERSION+" by Lymia")
    err.println("")
    args match {
      case Array(in) => 
        val file = new java.io.File(in)
        if(!file.exists()) err.println("File not found.")
        else if(!file.isFile()) err.println("Target is not a file.")
        else {
          Parser(new String(readFile(file))) match {
            case Left(expr) => compile(expr) match {
              case Left(expr) => println(expr)
              case Right(e) => err.println(e)
            }
            case Right(e) => err.println(e)
          }
        }
      case _ => err.println(USAGE)
    }
  }
  private def readFile(f: java.io.File): Array[Byte] = {
    import java.io._
    val data = new Array[Byte](f.length().asInstanceOf[Int])
    val in = new DataInputStream(new FileInputStream(f))
    in.readFully(data)
    in.close()
    return data
  }

  class MapExt[A,B](t: Map[A,B]) {
    def unsafeGet(k: A): B = t.getOrElse(k,sys.error("no such value"))
    def stringMap(f: ((A,B))=>Seq[Char]) = t.toSeq.stringMap(f)
  }
  implicit def mapExt[A,B](t: Map[A,B]) = new MapExt[A,B](t)
  class SeqExt[A](s: Seq[A]) {
    def stringMap(f: A=>Seq[Char]) = new String((s flatMap f).toArray)
    def stringMap(f: A=>Seq[Char], between: String, default: String) = 
      if(s.length>0) (s map {s => new String(f(s).toArray)}) reduceLeft {_+between+_}
      else default
  }
  implicit def seqExt[A](s: Seq[A]) = new SeqExt[A](s)

  val tq = "\"\"\""

  val prelude = """
    //Compiled by ulc on %s

    type Stack = scala.collection.immutable.List[Block]
    type Lambda = Stack=>Stack
    case class Block(f: Lambda, s: String)
    val Z = Block
    implicit def pushLiteral(u: Block) = (s: Stack)=>u::s
    class ComposeLambda(f: Lambda) { def ~(g: Lambda) = (x:Stack) => g(f(x)) }
    implicit def composeLambda(f: Lambda) = new ComposeLambda(f)
    implicit def composeLambda(f: Block) = new ComposeLambda(f)
    def exec(b: Block) {
      try {
        b.f(List())
      } catch {
        case e: NoSuchElementException => 
          println(" [ERROR: stack underflow]")
        case e: StackOverflowError => 
          println(" [ERROR: JVM stack overflow- try using recursion by fixed point combinator instead of ^]")
        case e => 
          println(" [ERROR: unexpected error "+e.getClass.getName+"]")
          e.printStackTrace()
      }
    }
    val a = (s: Stack)=>s
    val b = ""
    val c = "\"\"\""
    def a(s: String) = "("+s+")"
  """.replace("\n    ","\n").trim()
  val functionMappings = List(
    "~" -> """(s.tail.head) :: (s.head) :: (s.tail.tail)""",
    ":" -> """s.head :: s""",
    "!" -> """s.tail""",
    "*" -> """val a = s.tail.head; val b = s.head; Block(a.f~b.f,a.s+b.s) :: (s.tail.tail)""",
    "a" -> """val a = s.head; Block(a,"("+a.s+")") :: (s.tail)""",
    "^" -> """s.head.f(s.tail)""",
    "S" -> """print(s.head.s); s.tail"""
  )
  val nameToInternal = functionMappings.zipWithIndex.map{t => t._1._1 -> (('A'+t._2).toChar).toString}.toMap
  val codeBlock = functionMappings stringMap { t => "val %s=(s:Stack)=>{%s}\n".format(nameToInternal unsafeGet t._1, t._2) }
  val header = prelude+"\n"+codeBlock+"\n"

  val reservedWords = Set("abstract","case","catch","class","def","do","else","extends","false","final",
                          "finally","for","forSome","if","implicit","import","lazy","match","new","null",
                          "object","override","package","private","protected","return","sealed","super",
                          "this","throw","trait","try","true","type","val","var","while","with","yield")
  val symbolSeq = ('a' to 'z') ++ ('A' to 'Z')
  val symbolCount = symbolSeq.length
  def compile(s: Seq[Node]): Either[String,String] = {
    val b = new scala.collection.mutable.StringBuilder(header)
    var i = -1
    def genSym[T](a: T): (T, String) = {
      i = i+1
      var t = i
      var s = ""
      while(t >= symbolCount*symbolCount) {
        s+=symbolSeq(t%symbolCount)
        t/=symbolCount
      }
      val id = ""+symbolSeq(t/symbolCount)+symbolSeq(t%symbolCount)+s
      (a,if(reservedWords contains id) "_"+id
      else id)
    }

    val program = ParenNode(s)
    val parenList = parenNodes(program)
    val parenTextNames = parenList.map(genSym _).toMap
    val parenCodeNames = parenList.map(genSym _).toMap
    val written = new scala.collection.mutable.HashSet[Node]
    def compile(s: Node): Unit = s match {
      case x: ParenNode =>
        if(!written.contains(x)) {
          x.l map compile _
          b ++= "val %s=%s\n".format(parenTextNames unsafeGet x, (x.l) stringMap ( n => n match {
            case x: ParenNode => "a("+(parenTextNames unsafeGet x)+")"
            case TextNode(s)  => tq+s.replace(tq,tq+"+c+"+tq)+tq
          },"+","b"))
          b ++= "val %s=%s\n".format(parenCodeNames unsafeGet x, "Z("+(x.l stringMap (s => s match {
            case x: ParenNode => parenCodeNames unsafeGet x
            case x: TextNode  => compileNode(x.s)
          },"~","a"))+","+( parenTextNames unsafeGet x )+")")
          written += x
        }
      case TextNode(t) => Unit
    }
    compile(program)

    b ++= "exec("+(parenCodeNames unsafeGet program)+")"

    Left(b.toString)
  }
  def date = {
    import java.util._
    import java.text._
    val dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss")
    val date = new Date()
    dateFormat.format(date)
  }
  def parenNodes(s: ParenNode): Seq[ParenNode] = Seq(s) ++ (s.l flatMap { _ match {
    case x: ParenNode => parenNodes(x)
    case x: TextNode  => Seq()
  }})
  def compileNode(s: String) = (s:Seq[Char]) filter 
    {(s: Char) => nameToInternal contains s.toString} stringMap 
    ((s: Char) => nameToInternal unsafeGet s.toString,"~","a")

  sealed trait Node
  case class TextNode(s: String) extends Node {
    override lazy val toString=s
  }
  case class ParenNode(l: Seq[Node]) extends Node {
    override lazy val toString="("+(l stringMap {_.toString})+")"
  }

  object Parser extends scala.util.parsing.combinator.RegexParsers {
    override val skipWhitespace = false

    def block : Parser[Seq[Node]] = ((parens | string) *)
    def string: Parser[TextNode]  = "[^()]+".r ^^ TextNode
    def parens: Parser[ParenNode] = "(" ~> (block ^^ ParenNode) <~ ")"

    def apply(s:String): Either[Seq[Node],String] = parseAll(block, s) match {
      case Success(tree, _)    => Left(tree)
      case NoSuccess(err,next) => Right("At line "+next.pos.line+", column "+next.pos.column+": "+err)
    }
  }
}
